
#include <stdio.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#include "utils.h"
#include "control.h"
#include "definitions.h"


/****** Control public methods ******/

/** Control constructor
 *
 * A new object is created to control the various components of the program.
 * This object is fed a series of settings from a configuration, and
 * when Generate_VHDL is called, it produces the VHDL to represent the
 * processor.
 */
Control :: Control ( const char * opcode_map_file ,
                    const char * state_machine_directory ,
                    const char * root_vhdl_input_file ,
                    const char * vhdl_input_directory ,
                    const char * initial_state_machine ,
                    const char * required_opcode_list_file )
{
    /* First compile the regular expressions used in parsing the VHDL */
    int rc = regcomp ( & include_regex , 
        "^[\t ]*INCLUDE +([^ \t]+)[ \t]*" , REG_EXTENDED ) ;
    assert ( rc == 0 ) ;

    rc = regcomp ( & insert_subtypes_regex , 
        "^[\t ]*INSERT +SUBTYPES" , REG_EXTENDED ) ;
    assert ( rc == 0 ) ;

    rc = regcomp ( & insert_state_machine_regex , 
        "^[\t ]*INSERT +STATE +MACHINE" , REG_EXTENDED ) ;
    assert ( rc == 0 ) ;

    rc = regcomp ( & insert_instruction_decoder_regex , 
        "^[\t ]*INSERT +INSTRUCTION +DECODER" , REG_EXTENDED ) ;
    assert ( rc == 0 ) ;

    rc = regcomp ( & insert_optimisation_regex , 
        "^[\t ]*INSERT +OPTIMI[ZS]ATION +([^ \t]+)[ \t]*" , REG_EXTENDED ) ;
    assert ( rc == 0 ) ;

    /* This regex is used for parsing objdump/opcode list
     * files (Require_Opcodes_In_File) */
    rc = regcomp ( & require_opcode_regex , 
            require_opcode_expression , REG_EXTENDED | REG_ICASE ) ;
    assert ( rc == 0 ) ;


            
  
    opcode_database . Read_Opcode_Map ( opcode_map_file ) ;
    optimisation_manager . Read_Opcode_Map ( opcode_map_file ) ;

    /* Read in state machines */
    sm_loader . Add_State_Machine_Directory ( state_machine_directory ) ;

    Require_Opcodes_In_File ( required_opcode_list_file ) ;
    /* This will have to change: */
    //opcode_database . Require_Opcode ( 0x6000 ) ; // BRA
    //opcode_database . Require_Opcode ( 0x5280 ) ; // ADDQ
    //opcode_database . Require_Opcode ( 0x2001 ) ; // some move

    /* Use dependencies to calculate which microcode is needed to
     * run the program */
    sm_loader . Require_Microsubs 
                    ( opcode_database . List_Required_Micro_Subroutines () ) ;

    /* Build the master state machine for use as microcode. */
    master_sm = sm_loader . Build_Master_Machine ( initial_state_machine ) ;

    /* Finalise the opcode database, linking the instruction decoder
     * to the microcode */
    opcode_database . Finalise_DFA ( master_sm ) ;

    this -> vhdl_input_directory = Copy_String ( vhdl_input_directory ) ;
    this -> root_vhdl_input_file = Copy_String ( root_vhdl_input_file ) ;
}

/** Control destructor
 */
Control :: ~Control ()
{
    regfree ( & include_regex ) ; 
    regfree ( & insert_subtypes_regex ) ; 
    regfree ( & insert_state_machine_regex ) ; 
    regfree ( & insert_instruction_decoder_regex ) ;
    regfree ( & insert_optimisation_regex ) ;
    regfree ( & require_opcode_regex ) ;

    delete [] root_vhdl_input_file ;
    delete [] vhdl_input_directory ;
}

/** Generate_VHDL
 *
 * VHDL is generated and sent to the specified file.
 */
void Control :: Generate_VHDL ( FILE * output )
{
    Add_VHDL_Source_File ( root_vhdl_input_file , output ) ;
}
   
/****** Control private methods ******/

/** Add_VHDL_Source_File
 *
 * Recursively process the given VHDL source file, inserting the
 * appropriate information in the right places. This procedure
 * must be passed a finalised master state machine and opcode database.
 */
void Control :: Add_VHDL_Source_File ( const char * filename , FILE * output )
{
    FILE *      input ;
    
    regmatch_t  matches [ 3 ] ;
    char        str [ MAX_LINE_LEN + 1 ] ;

    /* First try to open the filename using the default path
     * provided to the constructor */
    char *      abs_filename = new char [ strlen ( vhdl_input_directory ) +
                        strlen ( filename ) + 2 ] ;
    strcpy ( abs_filename , vhdl_input_directory ) ;
    strcat ( abs_filename , "/" ) ;
    strcat ( abs_filename , filename ) ;
    input = fopen ( abs_filename , "rt" ) ;

    if ( input == 0L )
    {
        /* Now try reading it from the current directory */
        strcpy ( abs_filename , filename ) ;
        input = fopen ( abs_filename , "rt" ) ;
    }

    MESSAGE2 ( "Reading VHDL source '%s'.\n" , abs_filename ) ;

    if ( input == 0L )
    {
        throw new File_Access_Exception ( filename ) ;
    }


    while ( fgets ( str , MAX_LINE_LEN , input ) != NULL )
    {
        Remove_Trailing_Newlines ( str ) ;

        if ( regexec ( & include_regex , str , 2 , matches , 0 ) == 0 )
        {
            /* An include statement. Read the specified file. */
            char * new_filename = Get_Regex_Match ( str , & matches [ 1 ] ) ;

            Add_VHDL_Source_File ( new_filename , output ) ;

            delete [] new_filename ;
        } else if ( regexec ( & insert_subtypes_regex , str , 1 , 
                    matches , 0 ) == 0 )
        {
            /* This means that two subtype definitions should
            be added to the VHDL. They define state_register
            and stack_pointer_register */

            fprintf ( output ,
                "\tsubtype state_register is std_logic_vector "
                        "( %d downto 0 ) ;\n"
                "\tsubtype stack_pointer_register is std_logic_vector "
                        "( %d downto 0 ) ;\n" ,
                master_sm -> Get_Width_Of_State_Number () - 1 ,
                Get_Number_Of_Bits_Needed_For ( STACK_SIZE - 1 ) - 1 ) ;
        } else if ( regexec ( & insert_optimisation_regex , str , 2 , 
                    matches , 0 ) == 0 )
        {
            /* An optimisation statement. */
            char * ot = Get_Regex_Match ( str , & matches [ 1 ] ) ;

            optimisation_manager . Generate_VHDL ( output ,
                                Optimisation_Record ( ot ) ) ;

            delete [] ot ;
        } else if ( regexec ( & insert_state_machine_regex , str , 1 , 
                    matches , 0 ) == 0 )
        {
            /* This means that the state machine defintion
            should be added to the VHDL. */

            master_sm -> Compile_Machine ( output ) ;
        } else if ( regexec ( & insert_instruction_decoder_regex , str , 1 , 
                    matches , 0 ) == 0 )
        {
            /* This means that the instruction decoder 
            should be added to the VHDL. */

            opcode_database . Generate_VHDL ( output ) ;
        } else {
            /* The line will be added to the outgoing VHDL if it
            has any non-whitespace characters on it. */
            
            if ( String_Contains_Non_Whitespace ( str ) ) 
            {
                fputs ( str , output ) ;
                fputs ( "\n" , output ) ;
            }
        }
    }
    fclose ( input ) ;

    delete [] abs_filename ;
}

/** Require_Opcodes_In_File
 *
 * For each opcode listed in the file, run Require_Opcode.
 * The file may take one of the following formats:
 *    GNU objdump output
 *    a list of opcodes, one per line, in hex format.
 */
void Control :: Require_Opcodes_In_File ( const char * file )
{
    FILE *      input ;
    regmatch_t  matches [ 4 ] ;
    char        str [ MAX_LINE_LEN + 1 ] ;
    int         count = 0 ;
    int         line_no = 0 ;

    input = fopen ( file , "rt" ) ;

    if ( input == 0L )
    {
        throw new File_Access_Exception ( file ) ;
    }

    MESSAGE ( "Reading required opcodes list file '%s'\n" , file ) ;
    while ( fgets ( str , MAX_LINE_LEN , input ) != NULL )
    {
        line_no ++ ;

        Remove_Trailing_Newlines ( str ) ;
        Remove_Whitespace_From_Ends ( str ) ;

        if ( regexec ( & require_opcode_regex , str , 3 , matches , 0 ) == 0 )
        {
            /* Ah, a match. The 2nd field should be the opcode */
            char *  opcode_str = Get_Regex_Match ( str , & matches [ 2 ] ) ;
            int     opcode = strtol ( opcode_str , 0L , 16 ) ;

            opcode_database . Require_Opcode ( opcode ) ;

            optimisation_manager . Notify ( opcode ) ;

            count ++ ;

            delete [] opcode_str ;
        } else {
            MESSAGE2 ( "Unable to decode %s line %d: %s\n" ,
                        file , line_no , str ) ;
        }
    }
    fclose ( input ) ;

    MESSAGE ( "%d opcodes read.\n" , count ) ;
    if ( count == 0 )
    {
        throw new No_Opcodes_Read_Exception ( file ) ;
    }
}

const char * Control :: require_opcode_expression = 
                "^([0-9a-f]+:[\t ]+|0x|)([0-9a-f]{4})[ \t]" ;

