
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "optimisation.h"
#include "exceptions.h"
#include "utils.h"

    
/****** Optimisation_Record static data ******/
    
const Optimisation_Type_Data Optimisation_Record :: opt_data [] = { 
        { EA , "ea_mode" , "EM" } ,
        { EAREG , "ea_reg" , "em" } ,
        { ALUOP , "alu_internal_op" , "+-|&^c" } } ;


/****** Optimisation_Record public methods ******/


/** Optimisation_Record constructor
 *
 * A new optimisation record is created for the named optimisation type.
 */
Optimisation_Record :: Optimisation_Record ( const char * n )
{
    /* We don't know what subtype the optimisation will take.. */
    subtype = '\0' ;

    /* Search the opt_data table for the optimisation type */
    for ( int i = 0 ; i < NUMBER_OF_OPTIMISATIONS ; i ++ )
    {
        if ( strcasecmp ( n , opt_data [ i ] . name ) == 0 )
        {
            type = opt_data [ i ] . type ;
            return ;
        }
    }
    throw new Unrecognised_Optimisation_Exception ( n ) ;
}

/** Optimisation_Record constructor
 *
 * A new optimisation record is created for the named optimisation type.
 * Here, the name is specified as a code character. These are the same
 * characters that are used in the optimisation field of the opcode map.
 */
Optimisation_Record :: Optimisation_Record ( char c )
{
    /* Make a note of the optimisation's subtype */
    subtype = c ;

    /* Search the opt_data table for the optimisation type */
    for ( int i = 0 ; i < NUMBER_OF_OPTIMISATIONS ; i ++ )
    {
        if ( index ( opt_data [ i ] . codes , c ) != 0L )
        {
            type = opt_data [ i ] . type ;
            return ;
        }
    }
    char name [ 2 ] = { c , '\0' } ;

    throw new Unrecognised_Optimisation_Exception ( name ) ;
}

/** Print_Info
 *
 * Print the type name and subtype character of this optimisation record.
 */
void Optimisation_Record :: Print_Info ()
{
    MESSAGE3 ( "%s(%c) " , opt_data [ type ] . name , subtype ) ;
}


/****** Optimisation_NDFA_Accept_State public methods ******/

/** Optimisation_NDFA_Accept_State destructor
 *
 * The list of optimisation records is deleted.
 */
Optimisation_NDFA_Accept_State :: ~Optimisation_NDFA_Accept_State ()
{
    Optimisation_Record_Set_Iter        iter ;

    iter = opts . begin () ;
    while ( iter != opts . end () )
    {
        Optimisation_Record * item = (* iter) ;

        delete item ;

        iter ++ ;
    }
}


/****** Optimisation_Manager public methods ******/


/** Optimisation_Manager constructor
 *
 * The various optimisation classes are created.
 */
Optimisation_Manager :: Optimisation_Manager ()
                        : Opcode_Map_Reader ()
{
    optimisation_types [ EA ] = new EA_Optimisation () ;
    optimisation_types [ EAREG ] = new EA_Reg_Optimisation () ;
    optimisation_types [ ALUOP ] = new ALU_Optimisation () ;
}


/** Optimisation_Manager destructor
 *
 * The various optimisation classes are deleted.
 */
Optimisation_Manager :: ~Optimisation_Manager ()
{
    for ( int i = 0 ; i < NUMBER_OF_OPTIMISATIONS ; i ++ )
    {
        delete optimisation_types [ i ] ;
    }
}


/** Notify
 *
 * All the optimisation classes that are available for this opcode are
 * notified that it will appear in the program. They use this information
 * to work out what optimisations can be applied.
 */
void Optimisation_Manager :: Notify ( unsigned opcode )
{
    MESSAGE2 ( "Adding optimisation info for opcode 0x%04x..\n" , 
                        opcode ) ;

    Optimisation_NDFA_Accept_State *    accept_state =
        (Optimisation_NDFA_Accept_State *) ndfa . Get_Accept_State ( opcode ) ;

    if ( accept_state == 0L )
    {
        throw new UnavailableOpcodeException () ;
    }

    Optimisation_Record_Set             o_list = accept_state ->
                                            Get_Optimisation_Record_Set () ;


    Optimisation_Record_Set_Iter        iter ;

    iter = o_list . begin () ;
    while ( iter != o_list . end () )
    {
        Optimisation_Record * item = (* iter) ;

        optimisation_types [ item -> Get_Optimisation_Type () ] ->
                            Notify ( opcode , item -> Get_Subtype () ) ;
        iter ++ ;
    }
}



/** New_Accept_State
 *
 * Create a new accept state. This method is called from inherited method 
 * Read_Opcode_Map as data is read from the file.
 */
Accept_State * Optimisation_Manager :: New_Accept_State ( 
                const char * cmp_bit_pattern , 
                const char * opcode_name , 
                const char * dea , 
                const char * micro_sub_name , 
                const char * optimisation_info , 
                const char * comment )
{
    Optimisation_NDFA_Accept_State * accept_state ;

    /* Create the new accept state */
    accept_state = new Optimisation_NDFA_Accept_State ;

    /* Now, what optimisations should be present for this opcode? 
     * Check the optimisation_info field */

    MESSAGE3 ( "Adding optimisations for %s: " , opcode_name ) ;
    for ( int i = 0 ; i < (int) strlen ( optimisation_info ) ; i ++ )
    {
        if ( ! isspace ( optimisation_info [ i ] ) )
        {
            /* Create a new optimisation and add it to the accept state. */
            Optimisation_Record * new_opt =
                    new Optimisation_Record ( optimisation_info [ i ] ) ;

            new_opt -> Print_Info () ;

            accept_state -> Add_Optimisation ( new_opt ) ;
        }
    }

    MESSAGE3 ( "\n" ) ;

    return accept_state ;
}



/** Generate_VHDL
 *
 * The given optimisation type is asked for the VHDL that implements the
 * optimisation.
 */
void Optimisation_Manager :: Generate_VHDL ( 
                        FILE * output , Optimisation_Record o_type )
{
    optimisation_types [ o_type . Get_Optimisation_Type () ] ->
                                            Generate_VHDL ( output ) ;
}


